function getWindow(node) {
  var ownerDocument = node.ownerDocument;
  return ownerDocument ? ownerDocument.defaultView : window;
}

function getComputedStyle(element) {
  return getWindow(element).getComputedStyle(element);
}

var getElementMargins = (function (element) {
  // get the element margins, we need them to properly align the popper
  var styles = getComputedStyle(element);
  var top = parseFloat(styles.marginTop) || 0;
  var right = parseFloat(styles.marginRight) || 0;
  var bottom = parseFloat(styles.marginBottom) || 0;
  var left = parseFloat(styles.marginLeft) || 0;
  return {
    top: top,
    right: right,
    bottom: bottom,
    left: left
  };
});

var getElementClientRect = (function (element) {
  // get the basic client rect, it doesn't include margins
  var width = element.offsetWidth;
  var height = element.offsetHeight;
  var top = element.offsetTop;
  var left = element.offsetLeft;
  var margins = getElementMargins(element);
  return {
    width: width + margins.left + margins.right,
    height: height + margins.top + margins.bottom,
    y: top - margins.top,
    x: left - margins.left
  };
});

var getNodeName = (function (element) {
  return element ? element.nodeName.toUpperCase() : null;
});

var getParentNode = (function (element) {
  if (getNodeName(element) === 'HTML') {
    // DocumentElement detectedF
    return element;
  }

  return element.parentNode || // DOM Element detected
  // $FlowFixMe: need a better way to handle this...
  element.host || // ShadowRoot detected
  document.ownerDocument || // Fallback to ownerDocument if available
  document.documentElement // Or to documentElement if everything else fails
  ;
});

function getScrollParent(node) {
  if (!node) {
    return document.body;
  }

  if (['HTML', 'BODY', '#document'].includes(node.nodeName.toUpperCase())) {
    return node.ownerDocument.body;
  }

  if (node instanceof HTMLElement) {
    // Firefox want us to check `-x` and `-y` variations as well
    var _getComputedStyle = getComputedStyle(node),
        overflow = _getComputedStyle.overflow,
        overflowX = _getComputedStyle.overflowX,
        overflowY = _getComputedStyle.overflowY;

    if (/(auto|scroll|overlay)/.test(overflow + overflowY + overflowX)) {
      return node;
    }
  }

  return getScrollParent(getParentNode(node));
}

function listScrollParents(element, list) {
  if (list === void 0) {
    list = [];
  }

  var scrollParent = getScrollParent(element);
  var isBody = getNodeName(scrollParent) === 'BODY';
  var target = isBody ? getWindow(scrollParent) : scrollParent;
  var updatedList = list.concat(target);
  return isBody ? updatedList : updatedList.concat(listScrollParents(getParentNode(target)));
}

function getWindowScroll(node) {
  var win = getWindow(node);
  var scrollLeft = win.pageXOffset;
  var scrollTop = win.pageYOffset;
  return {
    scrollLeft: scrollLeft,
    scrollTop: scrollTop
  };
}

function getHTMLElementScroll(element) {
  return {
    scrollLeft: element.scrollLeft,
    scrollTop: element.scrollTop
  };
}

function getElementScroll(node) {
  if (node === getWindow(node) || !(node instanceof HTMLElement)) {
    return getWindowScroll(node);
  } else {
    return getHTMLElementScroll(node);
  }
}

function getOffsetParent(element) {
  var offsetParent = element instanceof HTMLElement ? element.offsetParent : null;
  var window = getWindow(element);

  if (getNodeName(offsetParent) === 'BODY') {
    return window;
  }

  return offsetParent || window;
}

var sumScroll = function sumScroll(scrollParents) {
  return scrollParents.reduce(function (scroll, scrollParent) {
    var nodeScroll = getElementScroll(scrollParent);
    scroll.scrollTop += nodeScroll.scrollTop;
    scroll.scrollLeft += nodeScroll.scrollLeft;
    return scroll;
  }, {
    scrollTop: 0,
    scrollLeft: 0
  });
};

function getCommonTotalScroll(reference, referenceScrollParents, popperScrollParents, limiter) {
  // if the scrollParent is shared between the two elements, we don't pick
  // it because it wouldn't add anything to the equation (they nulllify themselves)
  var nonCommonReference = referenceScrollParents.filter(function (node) {
    return !popperScrollParents.includes(node);
  }); // we then want to pick any scroll offset except for the one of the offsetParent
  // not sure why but that's how I got it working 😅
  // TODO: improve this comment with proper explanation

  var offsetParent = getOffsetParent(reference);
  var index = referenceScrollParents.findIndex(function (node) {
    return node === (limiter || offsetParent);
  });
  var scrollParents = referenceScrollParents.slice(0, index === -1 ? undefined : index);
  return sumScroll(scrollParents);
}

var unwrapJqueryElement = (function (element) {
  return element && element.jquery ? element[0] : element;
});

// source: https://stackoverflow.com/questions/49875255
var order = function order(modifiers) {
  // indexed by name
  var mapped = modifiers.reduce(function (mem, i) {
    mem[i.name] = i;
    return mem;
  }, {}); // inherit all dependencies for a given name

  var inherited = function inherited(i) {
    return mapped[i] && mapped[i].requires ? mapped[i].requires.reduce(function (mem, i) {
      return [].concat(mem, [i], inherited(i));
    }, []) : [];
  }; // order ...


  var ordered = modifiers.sort(function (a, b) {
    return !!~inherited(b.name).indexOf(a.name) ? -1 : 1;
  });
  return ordered;
};

var orderModifiers = (function (modifiers) {
  return [].concat(order(modifiers.filter(function (_ref) {
    var phase = _ref.phase;
    return phase === 'read';
  })), order(modifiers.filter(function (_ref2) {
    var phase = _ref2.phase;
    return phase === 'main';
  })), order(modifiers.filter(function (_ref3) {
    var phase = _ref3.phase;
    return phase === 'afterMain';
  })), order(modifiers.filter(function (_ref4) {
    var phase = _ref4.phase;
    return phase === 'write';
  })));
});

function expandToHashMap(value, keys) {
  return keys.reduce(function (hashMap, key) {
    hashMap[key] = value;
    return hashMap;
  }, {});
}

var getBasePlacement = (function (placement) {
  return placement.split('-')[0];
});

var getVariationPlacement = (function (placement) {
  return placement.split('-')[1];
});

var getMainAxisFromPlacement = (function (placement) {
  return ['top', 'bottom'].includes(placement) ? 'x' : 'y';
});

var top = 'top';
var bottom = 'bottom';
var right = 'right';
var left = 'left';
var basePlacements = [top, bottom, right, left];
var start = 'start';
var end = 'end';
var placements = basePlacements.reduce(function (acc, placement) {
  return acc.concat([placement, placement + "-" + start, placement + "-" + end]);
}, []); // modifiers that need to read the DOM

var read = 'read'; // pure-logic modifiers

var main = 'main'; // pure-logic modifiers that run after the main phase (such as computeStyles)

var write = 'write';

var computeOffsets = (function (_ref) {
  var reference = _ref.reference,
      element = _ref.element,
      strategy = _ref.strategy,
      placement = _ref.placement,
      scroll = _ref.scroll;
  var basePlacement = placement ? getBasePlacement(placement) : null;
  var variationPlacement = placement ? getVariationPlacement(placement) : null;
  var scrollTop = scroll.scrollTop,
      scrollLeft = scroll.scrollLeft;
  var offsets;

  switch (basePlacement) {
    case top:
      offsets = {
        x: reference.x + reference.width / 2 - element.width / 2 - scrollLeft,
        y: reference.y - element.height - scrollTop
      };
      break;

    case bottom:
      offsets = {
        x: reference.x + reference.width / 2 - element.width / 2 - scrollLeft,
        y: reference.y + reference.height - scrollTop
      };
      break;

    case right:
      offsets = {
        x: reference.x + reference.width - scrollLeft,
        y: reference.y + reference.height / 2 - element.height / 2 - scrollTop
      };
      break;

    case left:
      offsets = {
        x: reference.x - element.width - scrollLeft,
        y: reference.y + reference.height / 2 - element.height / 2 - scrollTop
      };
      break;

    default:
      offsets = {
        x: reference.x - scrollLeft,
        y: reference.y - scrollTop
      };
  }

  var mainAxis = basePlacement ? getMainAxisFromPlacement(basePlacement) : null;

  if (mainAxis != null) {
    var len = mainAxis === 'y' ? 'height' : 'width';

    switch (variationPlacement) {
      case start:
        offsets[mainAxis] -= reference[len] / 2 - element[len] / 2;
        break;

      case end:
        offsets[mainAxis] += reference[len] / 2 - element[len] / 2;
        break;
    }
  }

  return offsets;
});

var format = (function (str) {
  for (var _len = arguments.length, args = new Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
    args[_key - 1] = arguments[_key];
  }

  return [].concat(args).reduce(function (p, c) {
    return p.replace(/%s/, c);
  }, str);
});

function microtaskDebounce(fn) {
  var called = false;
  return function () {
    return new Promise(function (resolve) {
      if (called) {
        return resolve();
      }

      called = true;
      Promise.resolve().then(function () {
        called = false;
        resolve(fn());
      });
    });
  };
}

var ERROR_MESSAGE = 'Popper: modifier "%s" provided an invalid %s property, expected %s but got %s';
var VALID_PROPERTIES = ['name', 'enabled', 'phase', 'fn', 'onLoad', 'requires', 'options'];
var validateModifiers = (function (modifiers) {
  modifiers.forEach(function (modifier) {
    Object.keys(modifier).forEach(function (key) {
      switch (key) {
        case 'name':
          if (typeof modifier.name !== 'string') {
            console.error(format(ERROR_MESSAGE, String(modifier.name), '"name"', '"string"', "\"" + String(modifier.name) + "\""));
          }

          break;

        case 'enabled':
          if (typeof modifier.enabled !== 'boolean') {
            console.error(format(ERROR_MESSAGE, modifier.name, '"enabled"', '"boolean"', "\"" + String(modifier.enabled) + "\""));
          }

        case 'phase':
          if (![read, main, write].includes(modifier.phase)) {
            console.error(format(ERROR_MESSAGE, modifier.name, '"phase"', 'either "read", "main" or "write"', "\"" + String(modifier.phase) + "\""));
          }

          break;

        case 'fn':
          if (typeof modifier.fn !== 'function') {
            console.error(format(ERROR_MESSAGE, modifier.name, '"fn"', '"function"', "\"" + String(modifier.fn) + "\""));
          }

          break;

        case 'onLoad':
          if (typeof modifier.onLoad !== 'function') {
            console.error(format(ERROR_MESSAGE, modifier.name, '"onLoad"', '"function"', "\"" + String(modifier.fn) + "\""));
          }

          break;

        case 'requires':
          if (!Array.isArray(modifier.requires)) {
            console.error(format(ERROR_MESSAGE, modifier.name, '"requires"', '"array"', "\"" + String(modifier.requires) + "\""));
          }

          break;

        case 'options':
          break;

        default:
          console.error("PopperJS: an invalid property has been provided to the \"" + modifier.name + "\" modifier, valid properties are " + VALID_PROPERTIES.map(function (s) {
            return "\"" + s + "\"";
          }).join(', ') + "; but \"" + key + "\" was provided.");
      }
    });
  });
});

// DOM Utils
var INVALID_ELEMENT_ERROR = 'Popper: Invalid reference or popper argument provided to Popper, they must be either a valid DOM element or a jQuery-wrapped DOM element.';
var INFINITE_LOOP_ERROR = 'Popper: An infinite loop in the modifiers cycle has been detected! The cycle has been interrupted to prevent a browser crash.';

var areValidElements = function areValidElements(a, b) {
  return a instanceof Element && b instanceof Element;
};

var defaultOptions = {
  placement: 'bottom',
  eventListeners: {
    scroll: true,
    resize: true
  },
  modifiers: [],
  strategy: 'absolute'
};

var Popper =
/*#__PURE__*/
function () {
  function Popper(reference, popper, options) {
    var _this = this;

    if (options === void 0) {
      options = defaultOptions;
    }

    this.state = {
      placement: 'bottom',
      orderedModifiers: [],
      options: defaultOptions,
      modifiersData: {}
    };
    this.update = microtaskDebounce(function () {
      return new Promise(function (success, reject) {
        _this.forceUpdate();

        success(_this.state);
      });
    });
    // Unwrap `reference` and `popper` elements in case they are
    // wrapped by jQuery, otherwise consume them as is
    this.state.elements = {
      reference: unwrapJqueryElement(reference),
      popper: unwrapJqueryElement(popper)
    };
    var _this$state$elements = this.state.elements,
        referenceElement = _this$state$elements.reference,
        popperElement = _this$state$elements.popper; // Store options into state

    this.state.options = Object.assign({}, defaultOptions, {}, options); // Don't proceed if `reference` or `popper` are invalid elements

    if (!areValidElements(referenceElement, popperElement)) {
      if (process.env.NODE_ENV !== "production") {
        console.error(INVALID_ELEMENT_ERROR);
      }

      return;
    }

    this.state.scrollParents = {
      reference: listScrollParents(referenceElement),
      popper: listScrollParents(popperElement)
    }; // Order `options.modifiers` so that the dependencies are fulfilled
    // once the modifiers are executed

    this.state.orderedModifiers = orderModifiers([].concat(this.constructor.defaultModifiers, this.state.options.modifiers)) // Apply user defined preferences to modifiers
    .map(function (modifier) {
      return Object.assign({}, modifier, {}, _this.state.options.modifiers.find(function (_ref) {
        var name = _ref.name;
        return name === modifier.name;
      }));
    }); // Validate the provided modifiers so that the consumer will get warned
    // of one of the custom modifiers is invalid for any reason

    if (process.env.NODE_ENV !== "production") {
      validateModifiers(this.state.options.modifiers);
    } // Modifiers have the opportunity to execute some arbitrary code before
    // the first update cycle is ran, the order of execution will be the same
    // defined by the modifier dependencies directive.
    // The `onLoad` function may add or alter the options of themselves


    this.state.orderedModifiers.forEach(function (_ref2) {
      var onLoad = _ref2.onLoad,
          enabled = _ref2.enabled;
      return enabled && onLoad && onLoad(_this.state);
    });
    this.update().then(function () {
      // After the first update completed, enable the event listeners
      _this.enableEventListeners(_this.state.options.eventListeners);
    });
  } // Async and optimistically optimized update
  // it will not be executed if not necessary
  // debounced, so that it only runs at most once-per-tick


  var _proto = Popper.prototype;

  // Syncronous and forcefully executed update
  // it will always be executed even if not necessary, usually NOT needed
  // use Popper#update instead
  _proto.forceUpdate = function forceUpdate() {
    var _this2 = this;

    var _this$state$elements2 = this.state.elements,
        referenceElement = _this$state$elements2.reference,
        popperElement = _this$state$elements2.popper; // Don't proceed if `reference` or `popper` are not valid elements anymore

    if (!areValidElements(referenceElement, popperElement)) {
      if (process.env.NODE_ENV !== "production") {
        console.error(INVALID_ELEMENT_ERROR);
      }

      return;
    } // Get initial measurements
    // these are going to be used to compute the initial popper offsets
    // and as cache for any modifier that needs them later


    this.state.measures = {
      reference: getElementClientRect(referenceElement),
      popper: getElementClientRect(popperElement)
    }; // Get scrollTop and scrollLeft of the offsetParent
    // this will be used in the `computeOffsets` function to properly
    // position the popper taking in account the scroll position
    // FIXME: right now we only look for a single offsetParent (the popper one)
    //        but we really want to take in account the reference offsetParent as well

    var offsetParentScroll = getElementScroll(getOffsetParent(popperElement)); // Modifiers have the ability to read the current Popper state, included
    // the popper offsets, and modify it to address specifc cases

    this.state.reset = false; // Cache the placement in cache to make it available to the modifiers
    // modifiers will modify this one (rather than the one in options)

    this.state.placement = this.state.options.placement;
    this.state.orderedModifiers.forEach(function (modifier) {
      return _this2.state.modifiersData[modifier.name] = Object.assign({}, modifier.data);
    });
    var __debug_loops__ = 0;

    for (var index = 0; index < this.state.orderedModifiers.length; index++) {
      if (process.env.NODE_ENV !== "production") {
        __debug_loops__ += 1;

        if (__debug_loops__ > 100) {
          console.error(INFINITE_LOOP_ERROR);
          break;
        }
      }

      if (index === 0) {
        // Offsets are the actual position the popper needs to have to be
        // properly positioned near its reference element
        // This is the most basic placement, and will be adjusted by
        // the modifiers in the next step
        this.state.offsets = {
          popper: computeOffsets({
            reference: this.state.measures.reference,
            element: this.state.measures.popper,
            strategy: 'absolute',
            placement: this.state.placement,
            scroll: getCommonTotalScroll(referenceElement, this.state.scrollParents.reference, this.state.scrollParents.popper)
          })
        };
      }

      if (this.state.reset === true) {
        this.state.reset = false;
        index = -1;
        continue;
      }

      var _this$state$orderedMo = this.state.orderedModifiers[index],
          fn = _this$state$orderedMo.fn,
          enabled = _this$state$orderedMo.enabled,
          options = _this$state$orderedMo.options;

      if (enabled && typeof fn === 'function') {
        this.state = fn(this.state, options);
      }
    }
  };

  _proto.enableEventListeners = function enableEventListeners(eventListeners) {
    var _this3 = this;

    var _this$state$elements3 = this.state.elements,
        referenceElement = _this$state$elements3.reference,
        popperElement = _this$state$elements3.popper;

    var _ref3 = typeof eventListeners === 'boolean' ? expandToHashMap(eventListeners, ['scroll', 'resize']) : eventListeners,
        scroll = _ref3.scroll,
        resize = _ref3.resize;

    if (scroll) {
      var scrollParents = [].concat(this.state.scrollParents.reference, this.state.scrollParents.popper);
      scrollParents.forEach(function (scrollParent) {
        return scrollParent.addEventListener('scroll', _this3.update, {
          passive: true
        });
      });
    }

    if (resize) {
      var window = getWindow(this.state.elements.popper);
      window.addEventListener('resize', this.update, {
        passive: true
      });
    }
  };

  _proto.destroy = function destroy() {
    var _this4 = this;

    // Remove scroll event listeners
    var scrollParents = [].concat(this.state.scrollParents.reference, this.state.scrollParents.popper);
    scrollParents.forEach(function (scrollParent) {
      return scrollParent.removeEventListener('scroll', _this4.update);
    }); // Remove resize event listeners

    var window = getWindow(this.state.elements.popper);
    window.removeEventListener('resize', this.update); // Run `onDestroy` modifier methods

    this.state.orderedModifiers.forEach(function (_ref4) {
      var onDestroy = _ref4.onDestroy,
          enabled = _ref4.enabled;
      return enabled && onDestroy && onDestroy(_this4.state);
    });
  };

  return Popper;
}();
Popper.defaultModifiers = [];

export default Popper;
//# sourceMappingURL=popper-minimal.js.map
